#include <sys/param.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include "pathnames.h"

#define MAXLINE         1024

extern char *progname;
char *pathbuf, **arorder;

static FILE *cfp;

void cadd(char *add1, size_t len1, char *add2);
void enomem(void);

/*
 * getpath --
 *      read in the configuration file, calling a function with the line
 *      from each matching section.
 */
char *getpath(char **sects)
{
  char **av, *p;
  size_t len;
  char line[MAXLINE];
  static int openconfig(void);

  openconfig();
  while (fgets(line, sizeof(line), cfp))
  {
    if (!index(line, '\n'))
    {
      fprintf(stderr, "%s: config line too long.\n", progname);
      exit(1);
    }
    p = strtok(line, " \t\n");
    if (!p || *p == '#')
      continue;
    for (av = sects; *av; ++av)
      if (!strcmp(p, *av))
        break;
    if (!*av)
      continue;
    while (p = strtok((char *) NULL, " \t\n"))
    {
      len = strlen(p);
      if (p[len - 1] == '/')
        for (av = arorder; *av; ++av)
          cadd(p, len, *av);
      else
        cadd(p, len, (char *) NULL);
    }
  }
  return pathbuf;
}

void cadd(char *add1, size_t len1, char *add2)
{
  static size_t buflen;
  static char *bp, *endp;
  size_t len2;

  len2 = add2 ? strlen(add2) : 0;
  if (!bp || bp + len1 + len2 + 2 >= endp)
  {
    if (!(pathbuf = realloc(pathbuf, buflen += 1024)))
      enomem();
    if (!bp)
      bp = pathbuf;
    endp = pathbuf + buflen;
  }
  bcopy(add1, bp, len1);
  bp += len1;
  if (len2)
  {
    bcopy(add2, bp, len2);
    bp += len2;
  }
  *bp++ = ':';
  *bp = '\0';
}

static openconfig(void)
{
  if (cfp)
  {
    rewind(cfp);
    return;
  }
  if (!(cfp = fopen(_PATH_MANCONF, "r")))
  {
    fprintf(stderr, "%s: no configuration file %s.\n", progname, _PATH_MANCONF);
    exit(1);
  }
}

static char **_get_db_info(char *str)
{
  char *p;
  int cnt = -1, num = 0;
  char **ar = NULL, line[MAXLINE];

  openconfig();
  while (fgets(line, sizeof(line), cfp))
  {
    if (!index(line, '\n'))
    {
      fprintf(stderr, "%s: config line too long.\n", progname);
      exit(1);
    }
    p = strtok(line, " \t\n");
    if (!p || *p == '#' || strcmp(p, str))
      continue;
    while (p = strtok((char *) NULL, " \t\n"))
    {
      if (cnt == num - 1 && !(ar = realloc(ar, (num += 30) * sizeof(char **))))
          enomem();

      if (!(ar[++cnt] = strdup(p)))
        enomem();
    }
  }
  if (ar)
  {
    if (cnt == num - 1 && !(ar = realloc(ar, ++num * sizeof(char **))))
        enomem();

    ar[++cnt] = NULL;
  }
  return (ar);
}

char **getdb(void)
{
  return _get_db_info("_whatdb");
}

char **getorder(void)
{
  return _get_db_info("_subdir");
}

int getsection(char *sect)
{
  char *p;
  char line[MAXLINE];

  openconfig();
  while (fgets(line, sizeof(line), cfp))
  {
    if (!index(line, '\n'))
    {
      fprintf(stderr, "%s: config line too long.\n", progname);
      exit(1);
    }
    p = strtok(line, " \t\n");
    if (!p || *p == '#')
      continue;
    if (!strcmp(p, sect))
      return 1;
  }
  return 0;
}

void enomem(void)
{
  fprintf(stderr, "%s: %s\n", progname, strerror(ENOMEM));
  exit(1);
}
