#include <sys/param.h>
#include <sys/file.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "pathnames.h"

extern char **arorder, *pathbuf;

int f_all, f_cat, f_how, f_where;
int command = 0;
char *machine, *p_augment, *p_path, *pager, *progname;

static int arsize;
static int arpos;
static char **arnames;
static int debug;

void how(char *);
void cat(char *);
void add(char *);
void jump(char **, char *, char *);
void usage(void);
void do_command(void);
char *check_pager(char *);
char *getpath(char **);
char **getorder(void);

main(int argc, char **argv)
{
  extern char *optarg;
  extern int optind;
  int ch, res;
  char *section[2], *tmp;

  progname = "man";
  while ((ch = getopt(argc, argv, "-adcfhkM:m:P:w")) != EOF)
    switch (ch)
    {
      case 'a':
        f_all = 1;
        break;
      case 'c':
      case '-':         /* deprecated */
        f_cat = 1;
        break;
      case 'd':
        debug = 1;
        break;
      case 'h':
        f_how = 1;
        break;
      case 'm':
        p_augment = optarg;
        break;
      case 'M':
      case 'P':         /* backward compatibility */
        p_path = optarg;
        break;
        /*
         * "man -f" and "man -k" are backward compatible, undocumented
         * ways of calling whatis(1) and apropos(1).
         */
      case 'f':
        jump(argv, "-f", "whatis");
        /* NOTREACHED */
      case 'k':
        jump(argv, "-k", "apropos");
        /* NOTREACHED */
      case 'w':
        f_all = f_where = 1;
        break;
      case '?':
      default:
        usage();
    }
  argv += optind;

  if (!*argv)
    usage();

  if (!f_cat && !f_how)
    if (!isatty(1))
      f_cat = 1;
    else if (pager = getenv("PAGER"))
      pager = check_pager(pager);
    else
      pager = _PATH_PAGER;

  if (!(machine = getenv("MACHINE")))
    machine = MACHINE;

  /* see if checking in a specific section */
  if (argc > 1 && getsection(*argv))
  {
    section[0] = *argv++;
    section[1] = (char *) NULL;
  }
  else
  {
    section[0] = "_default";
    section[1] = (char *) NULL;
  }

  arorder = getorder();
  if (p_path || (p_path = getenv("MANPATH")))
  {
    char buf[MAXPATHLEN], **av;

    tmp = strtok(p_path, ":");
    while (tmp)
    {
      (void) sprintf(buf, "%s/", tmp);
      for (av = arorder; *av; ++av)
        cadd(buf, strlen(buf), *av);
      tmp = strtok((char *) NULL, ":");
    }
    p_path = pathbuf;
  }
  else if (!(p_path = getpath(section)) && !p_augment)
  {
    fprintf(stderr, "man: no place to search for those manual pages.\n");
    exit(1);
  }

  for (; *argv; ++argv)
  {
    if (p_augment)
      res = manual(p_augment, *argv);
    res = manual(p_path, *argv);
    if (res || f_where)
      continue;
    fprintf(stderr, "man: no entry for %s in the manual.\n", *argv);
    exit(1);
  }

  if (command)
    do_command();
  exit(0);
}

void complete_path(char *pathend, char *fname, char *suffix)
{
  strcat(pathend, "/");
  strcat(pathend, fname);
  if (suffix)
    strcat(pathend, suffix);
}

int found(char *path, char *fname, char *pathend, char *suffix)
{
  char *p;
  char command[MAXPATHLEN * 3];
  FILE *f;

  *pathend = '\0';
  complete_path(pathend, fname, suffix);
  if (debug)
    printf("Trying: '%s'\n", path);
  if (!access(path, R_OK))
    return 1;
  strcpy(pathend + 1, machine);
  complete_path(pathend, fname, suffix);
  if (debug)
    printf("Trying: '%s'\n", path);
  if (!access(path, R_OK))
    return 1;
  if (!suffix || suffix[0] != '.' || !strchr("0123456789l", suffix[1]) || suffix[2])
    return 0;
  for (p = pathend - 1; p >= path && *p != '/'; p--);
  if (*p++ != '/')
    return 0;
  if (strnicmp(p, "cat", 3))
    return 0;
  memcpy(p, "man", 3);
  *pathend = '\0';
  complete_path(pathend, fname, suffix);
  if (debug)
    printf("Trying: '%s'\n", path);
  if (access(path, R_OK))
  {
    strcpy(pathend + 1, machine);
    complete_path(pathend, fname, suffix);
    if (debug)
      printf("Trying: '%s'\n", path);
    if (access(path, R_OK))
    {
      memcpy(p, "cat", 3);
      return 0;
    }
  }
  f = fopen(path, "r");
  fread(command, 1, 5, f);
  fclose(f);
  if (!memcmp("'\\\" t", command, 5))
    strcpy(command, "cd /man ; groff -t -mandoc -Tascii ");
  else
    strcpy(command, "cd /man ; groff -mandoc -Tascii ");
  strcat(command, path);
  memcpy(p, "cat", 3);
  strcat(command, " >");
  p = path + strlen(path);
  if (strchr("123456789l", p[-1]))
    p[-1] = '0';
  strcat(command, path);
  fprintf(stderr, "please wait, formatting manual page `%s'\n", fname);
  system(command);
  return 1;
}

/*
 * manual --
 *      given a path, a directory list and a file name, find a file
 *      that matches; check ${directory}/${dir}/{file name} and
 *      ${directory}/${dir}/${machine}/${file name}.
 */
int manual(char *paths, char *name)
{
  int res = 0;
  char path[MAXPATHLEN + 1], *p, suf[3] = ".0";

  for (; *paths;)
  {
    for (p = path; *paths && *paths != ':'; *p++ = *paths++) ;
    if (*paths)
      paths++;
    if (p == path)
      continue;
    suf[1] = p[-1];
    if (found(path, name, p, ".0") ||
        found(path, name, p, suf) ||
        found(path, name, p, ".guide") ||
        found(path, name, p, ".dvi") ||
        found(path, name, p, ".doc") ||
        found(path, name, p, ".man") ||
        found(path, name, p, NULL))
    {
      if (f_where)
        printf("man: found in %s.\n", path);
      else if (f_cat)
        cat(path);
      else if (f_how)
        how(path);
      else
        add(path);
      if (!f_all)
        return 1;
      res = 1;
    }
  }
  return res;
}

/*
 * how --
 *      display how information
 */
void how(char *fname)
{
  FILE *fp;

  int lcnt, print;
  char *p;
  char buf[BUFSIZ];

  if (!(fp = fopen(fname, "r")))
  {
    fprintf(stderr, "man: %s: %s\n", fname, strerror(errno));
    exit(1);
  }
#define S1      "SYNOPSIS"
#define S2      "S\bSY\bYN\bNO\bOP\bPS\bSI\bIS\bS"
#define D1      "DESCRIPTION"
#define D2      "D\bDE\bES\bSC\bCR\bRI\bIP\bPT\bTI\bIO\bON\bN"
  for (lcnt = print = 0; fgets(buf, sizeof(buf), fp);)
  {
    if (!strncmp(buf, S1, sizeof(S1) - 1) || !strncmp(buf, S2, sizeof(S2) - 1))
    {
      print = 1;
      continue;
    }
    else if (!strncmp(buf, D1, sizeof(D1) - 1) || !strncmp(buf, D2, sizeof(D2) - 1))
      return;
    if (!print)
      continue;
    if (*buf == '\n')
      ++lcnt;
    else
    {
      for (; lcnt; --lcnt)
        putchar('\n');
      for (p = buf; isspace(*p); ++p);
      fputs(p, stdout);
    }
  }
  fclose(fp);
}

/*
 * cat --
 *      cat out the file
 */
void cat(char *fname)
{
  int n;
  FILE *f;
  char buf[BUFSIZ];

  if (!(f = fopen(fname, "r")))
  {
    fprintf(stderr, "man: %s: %s\n", fname, strerror(errno));
    exit(1);
  }
  while ((n = fread(buf, 1, sizeof(buf), f)) > 0)
  {
    if (fwrite(buf, 1, n, stdout) != n)
    {
      fprintf(stderr, "man: fwrite: %s\n", strerror(errno));
      exit(1);
    }
  }
  if (!feof(f))
  {
    fprintf(stderr, "man: fread: %s\n", strerror(errno));
    exit(1);
  }
  fclose(f);
}

/*
 * add --
 *      add a file name to the list for future paging
 */
void add(char *fname)
{
  if (arpos == arsize)
  {
    arsize += 10;
    if (!(arnames = realloc(arnames, arsize * sizeof(char *))))
      enomem();
  }
  if (!(arnames[arpos++] = strdup(fname)))
    enomem();
  if (*fname == '/')
  {
    fname = ++(arnames[arpos - 1]);
    for (; *fname && *fname != '/'; fname++);
    if (*fname)
      *fname = ':';
  }
  command = 1;
}

/*
 * check_pager --
 *      check the user supplied page information
 */
char *check_pager(char *name)
{
  char *p, *save;

  /*
   * if the user uses "more", we make it "more -s"; watch out for
   * PAGER = "mypager /usr/ucb/more"
   */
  for (p = name; *p && !isspace(*p); ++p);
  for (; p > name && *p != '/'; --p);
  if (p != name)
    ++p;

  /* make sure it's "more", not "morex" */
  if (!strncmp(p, "more", 4) && (!p[4] || isspace(p[4])))
  {
    save = name;
    /* allocate space to add the "-s" */
    if (!(name = malloc((u_int) (strlen(save) + sizeof("-s") + 1))))
      enomem();
    sprintf(name, "%s %s", save, "-s");
  }
  return name;
}

/*
 * jump --
 *      strip out flag argument and jump
 */
void jump(char **argv, char *flag, char *name)
{
  char **arg;

  argv[0] = name;
  for (arg = argv + 1; *arg; ++arg)
    if (!strcmp(*arg, flag))
      break;
  for (; *arg; ++arg)
    arg[0] = arg[1];
  execvp(name, argv);
  fprintf(stderr, "%s: Command not found.\n", name);
  exit(1);
}

void do_man_type(char *command, char *suffix, char *pager, int only_one)
{
  int i, sl, l, cnt = 0;
  
  if (!suffix)
    strcpy(command, pager);
  sl = suffix ? strlen(suffix) : 0;
  for (i = 0; i < arpos; i++)
    if (arnames[i])
    {
      l = strlen(arnames[i]);
      if (!sl || (l >= sl && !stricmp(arnames[i] + l - sl, suffix)))
      {
        if (only_one)
        {
          sprintf(command, pager, arnames[i]);
          system(command);
        }
        else
        {
          strcat(command, " ");
          strcat(command, arnames[i]);
          cnt++;
        }
        arnames[i] = NULL;
      }
    }
  if (!suffix && cnt)
    system(command);
}

void do_command(void)
{
  int i, len = strlen(pager) + 2;
  char *command;
  
  for (i = 0; i < arpos; len += strlen(arnames[i++]) + 1) ;
  if (!(command = malloc(len)))
    enomem();
  do_man_type(command, ".guide", "amigaguide %s", 1);
  do_man_type(command, ".dvi", "showdvi -l %s", 1);
  do_man_type(command, NULL, pager, 0);
}

/*
 * usage --
 *      print usage message and die
 */
void usage(void)
{
  fprintf(stderr, "usage: man [-ac] [-M path] [-m path] [section] title ...\n");
  exit(1);
}
